#define _XOPEN_SOURCE 500

#include <math.h>
#include <stdio.h>
#include <string.h>


#include "hardware/gpio.h"
#include "hardware/regs/rosc.h"
#include "hardware/regs/addressmap.h"

#include "pico/stdlib.h"

#include "external/fonts.h"
#include "external/st7735.h"
#include "external/ICM20948.h"

#include "chip8.h"
#include "peripherals.h"


void setup() {
  // Screen
  srand(time(NULL));
  stdio_init_all();
  setvbuf ( stdout , NULL , _IONBF , 0 );
  sleep_ms(1000);
  ST7735_Init();
  ST7735_DrawImage(0, 0, 80, 160, plaao_logo);
  sleep_ms(2000);
  printf("%s\n", "Initialising");

  // Accelerometer
  i2c_init(i2c0, 400 * 1000);
  gpio_set_function(4, GPIO_FUNC_I2C);
  gpio_set_function(5, GPIO_FUNC_I2C);
  gpio_pull_up(4);
  gpio_pull_up(5);
  sleep_ms(1000);
  
  IMU_EN_SENSOR_TYPE enMotionSensorType;

  imuInit(&enMotionSensorType);
  if (IMU_EN_SENSOR_TYPE_ICM20948 != enMotionSensorType) {
    printf("Failed to initialise IMU...");
  }

  printf("IMU initialised!\n");

  ST7735_FillScreen(ST7735_BLACK);
}

int main() {

    setup();

    printf("[PENDING] Initializing CHIP-8 arch...\n");
    init_cpu();
    printf("[OK] Done!\n");

 
    int status = load_rom();

    if (status == -1) {
        printf("[FAILED] fread() failure: the return value was not equal to the rom file size.\n");
        return 1;
    } else if (status != 0) {
        perror("Error while loading rom");
        return 1;
    }

    printf("[OK] Rom loaded successfully!\n");

    init_display();
    printf("[OK] Display successfully initialized.\n");
    int i = 0;

    float curX = 0;
    float curY = 0;
    float curZ = 0;

    float firstX = 0;
    float firstY = 0;
    float firstZ = 0;

    icm20948AccelRead(&firstX, &firstY, &firstZ);

    while (1) {

        icm20948AccelRead(&curX, &curY, &curZ);
        printf("x %f, y %f, z %f\n", curX, curY, curZ);
        float diffY = curY - firstY;
        if (diffY > 0.1) {
          keypad[7] = 1;
          keypad[9] = 0;
        } else if (diffY < -0.1) {
          keypad[9] = 1;
          keypad[7] = 0;
        } else {
          keypad[9] = 0;
          keypad[7] = 0;
        }

        emulate_cycle();
        if (draw_flag) {
          draw(display);
        }
    }
    return 0;
}